#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2020, Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Boilerplate
#-----------------------------------------------------------------------------
import logging # isort:skip
log = logging.getLogger(__name__)

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------

# Standard library imports
import itertools
import re
import warnings
from typing import (
    Any,
    Dict,
    Iterator,
    List,
    NamedTuple,
    Optional,
    Sequence,
    Tuple,
    Union,
)

# External imports
from typing_extensions import Literal

# Bokeh imports
from ..models import HoverTool, Plot, Tool, Toolbar
from ..models.tools import Drag, Inspection, Scroll, Tap

#-----------------------------------------------------------------------------
# Globals and constants
#-----------------------------------------------------------------------------

__all__ = (
    'process_active_tools',
    'process_tools_arg',
)

#-----------------------------------------------------------------------------
# General API
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Dev API
#-----------------------------------------------------------------------------

# TODO: str should be literal union of e.g. pan | xpan | ypan
Auto = Literal["auto"]
ActiveDrag = Union[Drag, Auto, str, None]
ActiveInspect = Union[List[Inspection], Inspection, Auto, str, None]
ActiveScroll = Union[Scroll, Auto, str, None]
ActiveTap = Union[Tap, Auto, str, None]

def process_active_tools(toolbar: Toolbar, tool_map: Dict[str, Tool],
        active_drag: ActiveDrag, active_inspect: ActiveInspect, active_scroll: ActiveScroll, active_tap: ActiveTap) -> None:
    """ Adds tools to the plot object

    Args:
        toolbar (Toolbar): instance of a Toolbar object
        tools_map (dict[str]): tool_map from _process_tools_arg
        active_drag (str or Tool): the tool to set active for drag
        active_inspect (str or Tool): the tool to set active for inspect
        active_scroll (str or Tool): the tool to set active for scroll
        active_tap (str or Tool): the tool to set active for tap

    Returns:
        None

    Note:
        This function sets properties on Toolbar
    """
    if active_drag in ['auto', None] or isinstance(active_drag, Tool):
        toolbar.active_drag = active_drag
    elif active_drag in tool_map:
        toolbar.active_drag = tool_map[active_drag]
    else:
        raise ValueError("Got unknown %r for 'active_drag', which was not a string supplied in 'tools' argument" % active_drag)

    if active_inspect in ["auto", None] or isinstance(active_inspect, Tool) or \
            (isinstance(active_inspect, list) and all(isinstance(t, Tool) for t in active_inspect)):
        toolbar.active_inspect = active_inspect
    elif isinstance(active_inspect, str) and active_inspect in tool_map:
        toolbar.active_inspect = tool_map[active_inspect]
    else:
        raise ValueError("Got unknown %r for 'active_inspect', which was not a string supplied in 'tools' argument" % active_scroll)

    if active_scroll in ['auto', None] or isinstance(active_scroll, Tool):
        toolbar.active_scroll = active_scroll
    elif active_scroll in tool_map:
        toolbar.active_scroll = tool_map[active_scroll]
    else:
        raise ValueError("Got unknown %r for 'active_scroll', which was not a string supplied in 'tools' argument" % active_scroll)

    if active_tap in ['auto', None] or isinstance(active_tap, Tool):
        toolbar.active_tap = active_tap
    elif active_tap in tool_map:
        toolbar.active_tap = tool_map[active_tap]
    else:
        raise ValueError("Got unknown %r for 'active_tap', which was not a string supplied in 'tools' argument" % active_tap)

def process_tools_arg(plot: Plot, tools: Union[str, Sequence[Union[Tool, str]]],
        tooltips: Optional[Union[str, Tuple[str, str]]] = None) -> Tuple[List[Tool], Dict[str, Tool]]:
    """ Adds tools to the plot object

    Args:
        plot (Plot): instance of a plot object

        tools (seq[Tool or str]|str): list of tool types or string listing the
            tool names. Those are converted using the to actual Tool instances.

        tooltips (string or seq[tuple[str, str]], optional):
            tooltips to use to configure a HoverTool

    Returns:
        list of Tools objects added to plot, map of supplied string names to tools
    """
    tool_objs, tool_map = _resolve_tools(tools)

    repeated_tools = [ str(obj) for obj in _collect_repeated_tools(tool_objs) ]
    if repeated_tools:
        warnings.warn("%s are being repeated" % ",".join(repeated_tools))

    if tooltips is not None:
        for tool_obj in tool_objs:
            if isinstance(tool_obj, HoverTool):
                tool_obj.tooltips = tooltips
                break
        else:
            tool_objs.append(HoverTool(tooltips=tooltips)) # type: ignore

    return tool_objs, tool_map

#-----------------------------------------------------------------------------
# Private API
#-----------------------------------------------------------------------------

def _resolve_tools(tools: Union[str, Sequence[Union[Tool, str]]]) -> Tuple[List[Tool], Dict[str, Tool]]:
    tool_objs = []
    tool_map = {}

    if not isinstance(tools, str):
        temp_tool_str = ""
        for tool in tools:
            if isinstance(tool, Tool):
                tool_objs.append(tool)
            elif isinstance(tool, str):
                temp_tool_str += tool + ','
            else:
                raise ValueError("tool should be a string or an instance of Tool class")
        tools = temp_tool_str

    for tool in re.split(r"\s*,\s*", tools.strip()):
        # re.split will return empty strings; ignore them.
        if tool == "":
            continue

        tool_obj = Tool.from_string(tool)
        tool_objs.append(tool_obj)
        tool_map[tool] = tool_obj

    return tool_objs, tool_map

def _collect_repeated_tools(tool_objs: List[Tool]) -> Iterator[Tool]:
    class Item(NamedTuple):
        obj: Tool
        properties: Dict[str, Any]

    key = lambda obj: obj.__class__.__name__

    for _, group in itertools.groupby(sorted(tool_objs, key=key), key=key):
        rest = [ Item(obj, obj.properties_with_values()) for obj in group ]
        while len(rest) > 1:
            head, *rest = rest
            for item in rest:
                if item.properties == head.properties:
                    yield item.obj

#-----------------------------------------------------------------------------
# Code
#-----------------------------------------------------------------------------



# -----------------------------------------------------------------------------
# Copyright (c) 2012 - 2020, Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------

# Standard library imports
from subprocess import CalledProcessError
from typing import Callable

# Bokeh imports
from .config import Config, StepType
from .enums import ActionStatus
from .system import cd, run
from .ui import failed, passed

__all__ = (
    "build_bokehjs",
    "build_conda_packages",
    "build_docs",
    "build_sdist_packages",
    "dev_install",
    "npm_install",
    "install_bokehjs",
)


def artifact_build_wrapper(name: str) -> Callable[[StepType], StepType]:
    def decorator(func: StepType) -> StepType:
        def wrapper(config: Config) -> None:
            try:
                config.build_status[name] = ActionStatus.STARTED
                func(config)
                passed(f"Build for {name!r} finished")
            except Exception as e:
                failed(f"Build for {name!r} did NOT succeed", str(e).split("\n"))
                config.abort()
            config.build_status[name] = ActionStatus.COMPLETED

        return wrapper

    return decorator


@artifact_build_wrapper("npm")
def build_bokehjs(config: Config) -> None:
    try:
        cd("bokehjs", silent=False)
        run("node make", silent=False)
        cd("..", silent=False)
        passed("BokehJS build succeeded")
    except CalledProcessError as e:
        failed("BokehJS build did NOT succeed", str(e).split("\n"))
        config.abort()


@artifact_build_wrapper("conda")
def build_conda_packages(config: Config) -> None:
    run("conda build conda.recipe --quiet --no-test", silent=False)


@artifact_build_wrapper("sdist")
def build_sdist_packages(config: Config) -> None:
    run("python setup.py sdist --install-js --formats=gztar", silent=False)


@artifact_build_wrapper("docs")
def build_docs(config: Config) -> None:
    cd("sphinx", silent=False)
    run("make clean all", silent=False, BOKEH_DOCS_CDN=config.version, BOKEH_DOCS_VERSION=config.version)
    cd("..", silent=False)


def npm_install(config: Config) -> None:
    try:
        cd("bokehjs", silent=False)
        run("npm install", silent=False)
        cd("..", silent=False)
        passed("npm install succeeded")
    except CalledProcessError as e:
        failed("npm install did NOT succeed", str(e).split("\n"))
        config.abort()


def install_bokehjs(config: Config) -> None:
    try:
        run("python setup.py --install-js", silent=False)
        passed("BokehJS install succeeded")
    except CalledProcessError as e:
        failed("BokehJS install did NOT succeed", str(e).split("\n"))
        config.abort()


def dev_install(config: Config) -> None:
    try:
        run("python setup.py develop --install-js")
        passed("Bokeh dev install succeeded")
    except CalledProcessError as e:
        failed("Bokeh dev install did NOT succeed", str(e).split("\n"))
        config.abort()



# -----------------------------------------------------------------------------
# Copyright (c) 2012 - 2020, Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------

# Standard library imports
import glob
from io import BytesIO
from subprocess import CalledProcessError
from typing import Any, Callable

# External imports
import boto
from packaging.version import Version as V

# Bokeh imports
from .config import Config, StepType
from .enums import ActionStatus
from .system import cd, run
from .ui import failed, passed, skipped

__all__ = (
    "publish_anaconda",
    "publish_cdn",
    "publish_docs",
    "publish_npm",
    "publish_pypi",
)


def upload_wrapper(name: str) -> Callable[[StepType], StepType]:
    def decorator(func: StepType) -> StepType:
        def wrapper(config: Config) -> None:
            try:
                config.upload_status[name] = ActionStatus.STARTED
                func(config)
                passed(f"Upload for {name!r} finished")
            except Exception as e:
                failed(f"Upload for {name!r} did NOT succeed", str(e).split("\n"))
                config.abort()
            config.upload_status[name] = ActionStatus.COMPLETED

        return wrapper

    return decorator


def upload_file_to_cdn(local_path: str, cdn_path: str, content_type: str, bucket: Any, binary: bool = False) -> None:
    print(":uploading to CDN: %s" % cdn_path)
    key = boto.s3.key.Key(bucket, cdn_path)
    key.metadata = {"Cache-Control": "max-age=31536000", "Content-Type": content_type}
    if binary:
        data = open(local_path, "rb").read()
    else:
        data = open(local_path).read().encode("utf-8")
    fp = BytesIO(data)
    key.set_contents_from_file(fp)


@upload_wrapper("cdn")
def publish_cdn(config: Config) -> None:
    subdir = "dev" if V(config.version).is_prerelease else "release"
    version = config.version

    access_key_id, secret_access_key = config.credentials["AWS"]

    buckets = []
    for bucket_name, bucket_region in [("cdn.bokeh.org", "us-east-1"), ("cdn-backup.bokeh.org", "us-west-2")]:
        conn = boto.s3.connect_to_region(
            bucket_region,
            aws_access_key_id=access_key_id,
            aws_secret_access_key=secret_access_key,
            calling_format=boto.s3.connection.OrdinaryCallingFormat(),
        )
        buckets.append(conn.get_bucket(bucket_name))

    content_type = "application/javascript"
    for name in ("bokeh", "bokeh-api", "bokeh-widgets", "bokeh-tables", "bokeh-gl"):
        for suffix in ("js", "min.js"):
            local_path = "bokehjs/build/js/%s.%s" % (name, suffix)
            cdn_path = "bokeh/%s/%s-%s.%s" % (subdir, name, version, suffix)
            for bucket in buckets:
                upload_file_to_cdn(local_path, cdn_path, content_type, bucket)


@upload_wrapper("anaconda")
def publish_anaconda(config: Config) -> None:
    if V(config.version).is_prerelease:
        cmd = "anaconda -t %s upload -u bokeh %s -l dev --force --no-progress"
    else:
        cmd = "anaconda -t %s upload -u bokeh %s -l dev -l main --force --no-progress"

    try:
        conda_base_dir = run("conda info --base").strip()
    except CalledProcessError as e:
        failed("Could not get conda base dir", str(e).split("\n"))
        config.abort()

    token = config.credentials["ANACONDA"]
    files = glob.glob(f"{conda_base_dir}/conda-bld/noarch/bokeh*.tar.bz2")
    for file in files:
        run(cmd % (token, file), fake_cmd=cmd % ("<hidden>", file), silent=False)

    files = glob.glob("dist/bokeh*.tar.gz")
    for file in files:
        run(cmd % (token, file), fake_cmd=cmd % ("<hidden>", file), silent=False)


@upload_wrapper("pypi")
def publish_pypi(config: Config) -> None:
    if V(config.version).is_prerelease:
        skipped("Not updating PyPI package for pre-releases")
        return

    token = config.credentials["PYPI"]
    cmd = "twine upload -u __token__ -p %s %s"
    files = glob.glob("dist/bokeh*.tar.gz")
    for file in files:
        run(cmd % (token, file), fake_cmd=cmd % ("<hidden>", file))


@upload_wrapper("docs")
def publish_docs(config: Config) -> None:
    cd("sphinx")
    sync_cmd = "aws s3 sync build/html s3://docs.bokeh.org/en/%s/ --acl bucket-owner-full-control --cache-control max-age=31536000,public"
    invalidate_cmd = "aws cloudfront create-invalidation --distribution-id E2OC6Q27H5UQ63 --paths %s"

    if V(config.version).is_prerelease:
        run(sync_cmd % "dev")
        run(invalidate_cmd % "/en/dev*")
    else:
        run(sync_cmd % config.version)
        run(sync_cmd % "latest")
        paths = "/en/latest* /versions.json"
        run(invalidate_cmd % paths)
    cd("..")


@upload_wrapper("npm")
def publish_npm(config: Config) -> None:
    if V(config.version).is_prerelease:
        skipped("Not updating NPM package for pre-releases")
        return

    cd("bokehjs")
    run("npm publish")
    cd("..")
